"""Easily convert your CALS, HTML, Formex 4, Office Open XML (docx) tables from one format to another."""

__version__ = "0.5.3"
__author__ = "Laurent LAPORTE <laurent.laporte.pro@gmail.com>"
__date__ = "2021-11-11"
__credit__ = "(c) 2018-2021 Laurent LAPORTE"
